# -*- coding: utf-8 -*-
# Copyright © 2023 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from collections import Iterable

from django.apps import apps
from django.conf import settings

from bridgeql.django.exceptions import InvalidBridgeQLSettings, InvalidAppOrModelName, InvalidModelFieldName
from bridgeql.utils import load_function
import os


DEFAULTS = {
    'BRIDGEQL_RESTRICTED_MODELS': {},
    'BRIDGEQL_AUTHENTICATION_DECORATOR': ''
}


class BridgeQLSettings:
    """
    Used to access bridgeql settings
    """

    def __init__(self):
        self.defaults = DEFAULTS

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid bridgeql setting '%s'" % attr)
        return getattr(settings, attr, self.defaults[attr])

    def _validate_restricted_models(self):
        """
        read settings.BRIDGEQL_RESTRICTED_MODELS and validate all models/fields
        BRIDGEQL_RESTRICTED_MODELS = {
            'auth.User': True,
            'machine.OperatingSystem': ['license_key'],
        }
        """
        restricted_models = self.BRIDGEQL_RESTRICTED_MODELS
        # restricted models in None
        if not restricted_models:
            return True

        if not isinstance(restricted_models, dict):
            raise InvalidBridgeQLSettings(
                'BRIDGEQL_RESTRICTED_MODELS requires dict value')

        # validate restricted fields
        for model in restricted_models.keys():
            try:
                app_name, model_name = str(model).split('.', 1)
                model_obj = apps.get_model(app_name, model_name)
                if isinstance(restricted_models[model], (list, tuple)):
                    for field_name in restricted_models[model]:
                        # fail fast in case attribute is not found
                        getattr(model_obj, field_name)
                elif not isinstance(restricted_models[model], bool):
                    raise AttributeError
            except (ValueError, LookupError):
                raise InvalidAppOrModelName('Invalid model %s in settings.BRIDGEQL_RESTRICTED_MODELS'
                                            % model)
            except AttributeError:
                raise InvalidModelFieldName(
                    'Invalid one or more fields %s in settings.BRIDGEQL_RESTRICTED_MODELS.'
                    % restricted_models[model])
        return True

    def _validate_auth_decorator(self):
        # check for the valid auth decorator
        if self.BRIDGEQL_AUTHENTICATION_DECORATOR:
            try:
                load_function(self.BRIDGEQL_AUTHENTICATION_DECORATOR)
            except (AttributeError, ImportError) as e:
                raise InvalidBridgeQLSettings(
                    'Wrong value for settings.BRIDGEQL_AUTHENTICATION_DECORATOR %s'
                    % self.BRIDGEQL_AUTHENTICATION_DECORATOR
                )
        return True

    def validate(self):
        return (
            self._validate_restricted_models() and
            self._validate_auth_decorator()
        )

    def get_local_apps(self):
        local_apps = []
        project_root = os.path.abspath(settings.BASE_DIR)
        for app in apps.get_app_configs():
            if os.path.dirname(app.path) == project_root:
                local_apps.append(app.label)
        return local_apps


bridgeql_settings = BridgeQLSettings()



# -*- coding: utf-8 -*-
# Copyright © 2023 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import json
from django.views.decorators.http import require_GET


from bridgeql.django.auth import auth_decorator
from bridgeql.django.exceptions import (
    ForbiddenModelOrField,
    InvalidRequest
)
from bridgeql.django.helpers import JSONResponse
from bridgeql.django.models import ModelBuilder

# TODO refine error handling


@auth_decorator
@require_GET
def read_django_model(request):
    params = request.GET.get('payload', None)
    try:
        params = json.loads(params)
        mb = ModelBuilder(params)
        qset = mb.queryset()  # get the result based on the given parameters
        res = {'data': qset, 'message': '', 'success': True}
        return JSONResponse(res)
    except ForbiddenModelOrField as e:
        res = {'data': [], 'message': str(e), 'success': False}
        return JSONResponse(res, status=403)
    except InvalidRequest as e:
        res = {'data': [], 'message': str(e), 'success': False}
        return JSONResponse(res, status=400)
    except Exception as e:
        res = {'data': [], 'message': str(e), 'success': False}
        return JSONResponse(res, status=500)

    """
    args = {
        selector: {
            and: [1,2,3],
            or: [1,2,3],
        },
        values: [],
        orderby: [],
        exclude: {
            buildid: 111
        },
        extras: {
            buildtree_url: []
        }
    }
    """



# -*- coding: utf-8 -*-
# Copyright © 2023 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause


class InvalidQueryException(Exception):
    pass


class ForbiddenModelOrField(Exception):
    pass


class InvalidRequest(Exception):
    pass


class InvalidAppOrModelName(InvalidRequest):
    pass


class InvalidModelFieldName(InvalidRequest):
    pass


class InvalidBridgeQLSettings(Exception):
    pass



# -*- coding: utf-8 -*-
# Copyright © 2023 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from django.db.models import Q

from bridgeql.django.exceptions import InvalidQueryException


def construct_query(selector):
    """
    selector: {
        'pk': 123,
        'type': 'release',
        '__or': [
            {'user': args.user},
            {'email': args.email}
        ]
    }

    selector: {
        '__or': [
            {'pk': 1, 'user': 'John Doe'},
            {'buildtype': 'release'}
        ]
    }
    """
    query = Q()
    if selector is None:
        return query
    if not isinstance(selector, dict):
        raise InvalidQueryException(
            "Selector is of the type %s, expected dict" % type(selector))
    or_filter = selector.pop('__or', [])
    if or_filter:
        if not isinstance(or_filter, list):
            raise InvalidQueryException(
                "__or filter is of the type %s, expected list" % type(or_filter))
        for f in or_filter:
            query.add(construct_query(f), conn_type=Q.OR)
    return query & Q(**dict(selector))


def extract_keys(query_dict):
    exclusion = ('__or',)
    keys = []
    for key, val in query_dict.items():
        if key not in exclusion:
            keys.append(key)
        else:
            for nested_dict in val:
                keys.extend(extract_keys(nested_dict))
    return keys



# -*- coding: utf-8 -*-
# Copyright © 2023 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from django.apps import apps
from django.core.exceptions import FieldDoesNotExist
from django.db.models import QuerySet
from django.db.models.base import ModelBase

from bridgeql.django import logger
from bridgeql.django.exceptions import (
    ForbiddenModelOrField,
    InvalidRequest,
    InvalidAppOrModelName,
    InvalidModelFieldName
)
from bridgeql.django.query import construct_query, extract_keys
from bridgeql.django.settings import bridgeql_settings


class Parameters(object):
    def __init__(self, params):
        self.params = params
        self.db_name = None  # db to connect
        self.app_name = None
        self.model_name = None
        self.filter = {}
        self.exclude = {}
        self.fields = []
        self.order_by = []
        self.distinct = False
        self.count = False
        self.limit = None
        self.offset = 0  # default offset is 0
        self._inject_params()

    def _inject_params(self):
        """
        populate all ModelBuilder object variables and validate it
        """
        for param, value in self.params.items():
            setattr(self, param, value)
        # validation for required fields
        if not any((self.app_name, self.model_name)):
            raise InvalidRequest('app_name or model_name missing')


class DBRows(list):
    # override count method of list
    def count(self):
        return len(self)


class Field(object):
    def __init__(self, model_config, field_name):
        self.model_config = model_config
        self.name = field_name
        self._resolve_pk()

    def _resolve_pk(self):
        if self.name == 'pk':
            self.name = self.model_config.model._meta.pk.name

    @property
    def is_restricted(self):
        return (self.name in self.model_config.restricted_fields)


class ModelConfig(object):
    def __init__(self, app_name, model_name):
        self.app_name = app_name
        self.model_name = model_name
        self.restricted_fields = self._get_restricted_fields()
        self.model = self._get_model()  # restricted_fields list to set
        self.fields = self._get_fields()

    def _get_fields(self):
        return set([f.name for f in self.model._meta.get_fields()])

    def get_restricted_fields_in(self, fields):
        return self.restricted_fields.intersection(set(fields))

    def is_restricted(self, fields):
        return bool(self.get_restricted_fields_in(fields))

    def _get_restricted_fields(self):
        # get from settings
        restricted_fields = bridgeql_settings.BRIDGEQL_RESTRICTED_MODELS.get(
            self.full_model_name, [])
        if isinstance(restricted_fields, list):
            restricted_fields = set(restricted_fields)
        return restricted_fields

    def _get_model(self):
        if self.full_model_name in bridgeql_settings.BRIDGEQL_RESTRICTED_MODELS:
            if self.restricted_fields is True:
                raise ForbiddenModelOrField(
                    'Unable to access restricted model %s.' % self.full_model_name)
        try:
            model = apps.get_model(self.app_name,
                                   self.model_name)
        except LookupError:
            raise InvalidAppOrModelName(
                'Invalid app or model name %s.' % self.full_model_name)
        return model

    @property
    def full_model_name(self):
        return "%s.%s" % (
            self.app_name,
            self.model_name
        )

    def validate_fields(self, query_fields):
        # combination of all fields used in query
        for q_field in query_fields:
            parent = self
            for field in q_field.split('__'):
                field_obj = Field(parent, field)
                if field_obj.is_restricted:
                    raise ForbiddenModelOrField('%s is restricted for model %s' %
                                                (field_obj.name, parent.full_model_name))
                try:
                    attr = parent.model._meta.get_field(
                        field_obj.name).related_model
                    if isinstance(attr, ModelBase):
                        parent = ModelConfig(
                            attr._meta.app_label, attr._meta.object_name)
                    else:
                        break
                except FieldDoesNotExist:
                    raise InvalidModelFieldName(
                        'Invalid field name %s for model %s.' %
                        (field_obj.name, parent.full_model_name)
                    )
        return True


class ModelBuilder(object):
    _QUERYSET_OPTS = [
        ('exclude', 'exclude'),  # dict
        ('distinct', 'distinct'),  # bool
        ('order_by', 'order_by'),  # list
        ('fields', 'values'),  # list
        ('count', 'count'),  # bool
    ]

    def __init__(self, params):
        self.params = Parameters(params)
        self.qset = None

        self.model_config = ModelConfig(
            self.params.app_name, self.params.model_name)
        requested_fields = list()
        requested_fields.extend(extract_keys(self.params.filter))
        requested_fields.extend(extract_keys(self.params.exclude))
        requested_fields.extend(self.params.fields)
        requested_fields.extend(self.params.order_by)
        self.model_config.validate_fields(set(requested_fields))

    def _apply_opts(self):
        for opt, qset_opt in ModelBuilder._QUERYSET_OPTS:
            func = getattr(self.qset, qset_opt)
            value = getattr(self.params, opt, None)
            if not value:
                continue
            if isinstance(value, dict):
                self.qset = func(**value)
            elif isinstance(value, list):
                # handle values case where property is passed in fields
                if qset_opt == 'values' and self.query_has_properties():
                    # returns DBRows instance
                    self.qset = self._add_fields()
                else:
                    self.qset = func(*value)
            else:
                self.qset = func()

    def query_has_properties(self):
        # TODO show error if distinct is True and properties are present in fields
        if self.params.distinct:
            return False
        return bool(set(self.params.fields) - self.model_config.fields)

    def _add_fields(self):
        qset_values = DBRows()
        self.print_db_query_log()
        for row in self.qset:
            model_fields = {}
            for field in self.params.fields:
                attr = row
                for ref in field.split('__'):
                    try:
                        attr = getattr(attr, ref)
                    except AttributeError:
                        raise InvalidModelFieldName(
                            'Invalid query for field %s.' % ref)
                model_fields[field] = attr
            qset_values.append(model_fields)
        return qset_values

    def queryset(self):
        # construct Q object from dictionary
        # x = Machine.objects.filter(name__startswith='machine-name-1')
        # x.distinct().order_by('os__name').values('os__name','ip').count()
        query = construct_query(self.params.filter)
        if self.params.db_name:
            self.qset = self.model_config.model.objects.using(
                self.params.db_name).filter(query)
        else:
            self.qset = self.model_config.model.objects.filter(query)
        self._apply_opts()
        # handle limit and offset seperately
        if self.params.limit:
            self.qset = self.qset[self.params.offset:
                                  self.params.offset + self.params.limit]
        if isinstance(self.qset, QuerySet):
            self.print_db_query_log()
            return list(self.qset)
        return self.qset

    def print_db_query_log(self):
        logger.debug('Request parameters: %s \nQuery: %s',
                     self.params.params, self.qset.query)

# query -> normal fields, foreign key reference



# -*- coding: utf-8 -*-
# Copyright © 2023 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

try:
    from django.urls import path
except ImportError:
    from django.conf.urls import url as path

from bridgeql.django.bridge import read_django_model
from bridgeql.django.settings import bridgeql_settings

bridgeql_settings.validate()

urlpatterns = [
    path('dj_read/', read_django_model, name='bridgeql_django_read'),
]



# -*- coding: utf-8 -*-
# Copyright © 2023 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from datetime import datetime
import json

from django.http import HttpResponse
from django.db.models.query import QuerySet


class JSONEncoder(json.JSONEncoder):
    '''
    Encode an object in JSON.
    '''

    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.ctime()
        if hasattr(obj, '__json__'):
            return obj.__json__()
        return json.JSONEncoder.default(self, obj)


class JSONResponse(HttpResponse):
    '''
    Create a response that contains a JSON string.
    '''

    def __init__(self, content,
                 content_type='application/json; charset=utf-8', status=200,
                 encoder=JSONEncoder):
        if isinstance(content, QuerySet):
            # evaluate the queryset with list(content)
            data = json.dumps(list(content), indent=3, cls=encoder)
        else:
            data = json.dumps(content, indent=3, cls=encoder)
        HttpResponse.__init__(self, data, content_type, status)



# -*- coding: utf-8 -*-
# Copyright © 2023 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

from django.http import HttpResponse
from django.contrib.auth import authenticate

from bridgeql.django.settings import bridgeql_settings
from bridgeql.utils import b64decode, load_function


def basic_auth(api):
    def wrap(request, *args, **kwargs):
        auth = request.META.get('HTTP_AUTHORIZATION', '')
        if auth and len(auth.split()) == 2:
            if auth[0].lower() == "basic":
                uname, passwd = b64decode(auth[1]).split(':', 1)
                user = authenticate(username=uname, password=passwd)
                if user is not None and user.is_active:
                    request.user = user
                    return api(request, *args, **kwargs)
        response = HttpResponse(status=401)
        response['WWW-Authenticate'] = 'Basic base64(user:password)'
        return response
    return wrap


if bridgeql_settings.BRIDGEQL_AUTHENTICATION_DECORATOR:
    auth_decorator = load_function(
        bridgeql_settings.BRIDGEQL_AUTHENTICATION_DECORATOR)
else:
    def auth_decorator(func): return func



# -*- coding: utf-8 -*-
# Copyright © 2023 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="bridgeql",
    version="0.1.10",
    author="Piyus Kumar",
    author_email="piyusk@vmware.com",
    description="Query language to bridge the gap between REST API and ORM capability",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='BSD-2',
    platforms=['Any'],
    keywords=['django'],
    url="https://github.com/vmware/bridgeql",
    project_urls={
        "Bug Tracker": "https://github.com/vmware/bridgeql/issues"
    },
    classifiers={
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Framework :: Django",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.1",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
    },
    packages=[
        "bridgeql",
        "bridgeql.django",
    ],
    python_requires=">=2.7"
)



import base64
import importlib
import json
import socket
import sys

PY_VERSION = sys.version_info.major


def get_client_ip(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


def local_ip_hostname():
    hostname = socket.getfqdn()
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('192.255.255.255', 1))
        ip_address = s.getsockname()[0]
    except:
        ip_address = '127.0.0.1'
    finally:
        s.close()
    return ip_address, hostname


def b64encode(data):
    if PY_VERSION >= 3:
        return base64.b64encode(data.encode('utf-8'))
    return base64.b64encode(data)


def b64decode(data):
    if PY_VERSION >= 3:
        return base64.b64decode(data).decode('utf-8')
    return base64.b64decode(data)


def b64encode_json(data):
    if PY_VERSION >= 3:
        return base64.b64encode(json.dumps(data).encode('utf-8'))
    return base64.b64encode(json.dumps(data))


def b64decode_json(data):
    if PY_VERSION >= 3:
        return json.loads(base64.b64decode(data).decode('utf-8'))
    return json.loads(base64.b64decode(data))


def load_function(function_str):
    mod_name, func_name = function_str.rsplit('.', 1)
    mod = importlib.import_module(mod_name)
    return getattr(mod, func_name)
